@@ -1,9 +1,19 @@
+from __future__ import annotations
+
 import enum
 import itertools
 import socket
 import sys
 from contextlib import contextmanager
-from typing import TYPE_CHECKING
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Iterator,
+    Literal,
+    TypeVar,
+    cast,
+)
 
 import attr
 from outcome import Value
@@ -14,12 +24,15 @@ from ._run import _public
 from ._windows_cffi import (
     INVALID_HANDLE_VALUE,
     AFDPollFlags,
+    CData,
     CompletionModes,
     ErrorCodes,
     FileFlags,
+    Handle,
     IoControlCodes,
     WSAIoctls,
     _handle,
+    _Overlapped,
     ffi,
     kernel32,
     ntdll,
@@ -27,7 +40,15 @@ from ._windows_cffi import (
     ws2_32,
 )
 
-assert not TYPE_CHECKING or sys.platform == "win32"
+if TYPE_CHECKING:
+    from typing_extensions import Buffer, TypeAlias
+
+    from .._file_io import _HasFileNo
+    from ._traps import Abort, RaiseCancelT
+    from ._unbounded_queue import UnboundedQueue
+
+EventResult: TypeAlias = int
+T = TypeVar("T")
 
 # There's a lot to be said about the overall design of a Windows event
 # loop. See
@@ -179,113 +200,6 @@ class CKeys(enum.IntEnum):
     USER_DEFINED = 4  # and above
 
 
-def _check(success):
-    if not success:
-        raise_winerror()
-    return success
-
-
-def _get_underlying_socket(sock, *, which=WSAIoctls.SIO_BASE_HANDLE):
-    if hasattr(sock, "fileno"):
-        sock = sock.fileno()
-    base_ptr = ffi.new("HANDLE *")
-    out_size = ffi.new("DWORD *")
-    failed = ws2_32.WSAIoctl(
-        ffi.cast("SOCKET", sock),
-        which,
-        ffi.NULL,
-        0,
-        base_ptr,
-        ffi.sizeof("HANDLE"),
-        out_size,
-        ffi.NULL,
-        ffi.NULL,
-    )
-    if failed:
-        code = ws2_32.WSAGetLastError()
-        raise_winerror(code)
-    return base_ptr[0]
-
-
-def _get_base_socket(sock):
-    # There is a development kit for LSPs called Komodia Redirector.
-    # It does some unusual (some might say evil) things like intercepting
-    # SIO_BASE_HANDLE (fails) and SIO_BSP_HANDLE_SELECT (returns the same
-    # socket) in a misguided attempt to prevent bypassing it. It's been used
-    # in malware including the infamous Lenovo Superfish incident from 2015,
-    # but unfortunately is also used in some legitimate products such as
-    # parental control tools and Astrill VPN. Komodia happens to not
-    # block SIO_BSP_HANDLE_POLL, so we'll try SIO_BASE_HANDLE and fall back
-    # to SIO_BSP_HANDLE_POLL if it doesn't work.
-    # References:
-    # - https://github.com/piscisaureus/wepoll/blob/0598a791bf9cbbf480793d778930fc635b044980/wepoll.c#L2223
-    # - https://github.com/tokio-rs/mio/issues/1314
-
-    while True:
-        try:
-            # If this is not a Komodia-intercepted socket, we can just use
-            # SIO_BASE_HANDLE.
-            return _get_underlying_socket(sock)
-        except OSError as ex:
-            if ex.winerror == ErrorCodes.ERROR_NOT_SOCKET:
-                # SIO_BASE_HANDLE might fail even without LSP intervention,
-                # if we get something that's not a socket.
-                raise
-            if hasattr(sock, "fileno"):
-                sock = sock.fileno()
-            sock = _handle(sock)
-            next_sock = _get_underlying_socket(
-                sock, which=WSAIoctls.SIO_BSP_HANDLE_POLL
-            )
-            if next_sock == sock:
-                # If BSP_HANDLE_POLL returns the same socket we already had,
-                # then there's no layering going on and we need to fail
-                # to prevent an infinite loop.
-                raise RuntimeError(
-                    "Unexpected network configuration detected: "
-                    "SIO_BASE_HANDLE failed and SIO_BSP_HANDLE_POLL didn't "
-                    "return a different socket. Please file a bug at "
-                    "https://github.com/python-trio/trio/issues/new, "
-                    "and include the output of running: "
-                    "netsh winsock show catalog"
-                )
-            # Otherwise we've gotten at least one layer deeper, so
-            # loop back around to keep digging.
-            sock = next_sock
-
-
-def _afd_helper_handle():
-    # The "AFD" driver is exposed at the NT path "\Device\Afd". We're using
-    # the Win32 CreateFile, though, so we have to pass a Win32 path. \\.\ is
-    # how Win32 refers to the NT \GLOBAL??\ directory, and GLOBALROOT is a
-    # symlink inside that directory that points to the root of the NT path
-    # system. So by sticking that in front of the NT path, we get a Win32
-    # path. Alternatively, we could use NtCreateFile directly, since it takes
-    # an NT path. But we already wrap CreateFileW so this was easier.
-    # References:
-    #   https://blogs.msdn.microsoft.com/jeremykuhne/2016/05/02/dos-to-nt-a-paths-journey/
-    #   https://stackoverflow.com/a/21704022
-    #
-    # I'm actually not sure what the \Trio part at the end of the path does.
-    # Wepoll uses \Device\Afd\Wepoll, so I just copied them. (I'm guessing it
-    # might be visible in some debug tools, and is otherwise arbitrary?)
-    rawname = r"\\.\GLOBALROOT\Device\Afd\Trio".encode("utf-16le") + b"\0\0"
-    rawname_buf = ffi.from_buffer(rawname)
-
-    handle = kernel32.CreateFileW(
-        ffi.cast("LPCWSTR", rawname_buf),
-        FileFlags.SYNCHRONIZE,
-        FileFlags.FILE_SHARE_READ | FileFlags.FILE_SHARE_WRITE,
-        ffi.NULL,  # no security attributes
-        FileFlags.OPEN_EXISTING,
-        FileFlags.FILE_FLAG_OVERLAPPED,
-        ffi.NULL,  # no template file
-    )
-    if handle == INVALID_HANDLE_VALUE:  # pragma: no cover
-        raise_winerror()
-    return handle
-
-
 # AFD_POLL has a finer-grained set of events than other APIs. We collapse them
 # down into Unix-style "readable" and "writable".
 #
@@ -330,9 +244,9 @@ WRITABLE_FLAGS = (
 # operation and start a new one.
 @attr.s(slots=True, eq=False)
 class AFDWaiters:
-    read_task = attr.ib(default=None)
-    write_task = attr.ib(default=None)
-    current_op = attr.ib(default=None)
+    read_task: _core.Task | None = attr.ib(default=None)
+    write_task: _core.Task | None = attr.ib(default=None)
+    current_op: AFDPollOp | None = attr.ib(default=None)
 
 
 # We also need to bundle up all the info for a single op into a standalone
@@ -340,10 +254,10 @@ class AFDWaiters:
 # finishes, even if we're throwing it away.
 @attr.s(slots=True, eq=False, frozen=True)
 class AFDPollOp:
-    lpOverlapped = attr.ib()
-    poll_info = attr.ib()
-    waiters = attr.ib()
-    afd_group = attr.ib()
+    lpOverlapped: CData = attr.ib()
+    poll_info: Any = attr.ib()
+    waiters: AFDWaiters = attr.ib()
+    afd_group: AFDGroup = attr.ib()
 
 
 # The Windows kernel has a weird issue when using AFD handles. If you have N
@@ -359,17 +273,20 @@ MAX_AFD_GROUP_SIZE = 500  # at 1000, the cubic scaling is just starting to bite
 
 @attr.s(slots=True, eq=False)
 class AFDGroup:
-    size = attr.ib()
-    handle = attr.ib()
+    size: int = attr.ib()
+    handle: Handle = attr.ib()
+
+
+assert not TYPE_CHECKING or sys.platform == "win32"
 
 
 @attr.s(slots=True, eq=False, frozen=True)
 class _WindowsStatistics:
-    tasks_waiting_read = attr.ib()
-    tasks_waiting_write = attr.ib()
-    tasks_waiting_overlapped = attr.ib()
-    completion_key_monitors = attr.ib()
-    backend = attr.ib(default="windows")
+    tasks_waiting_read: int = attr.ib()
+    tasks_waiting_write: int = attr.ib()
+    tasks_waiting_overlapped: int = attr.ib()
+    completion_key_monitors: int = attr.ib()
+    backend: Literal["windows"] = attr.ib(init=False, default="windows")
 
 
 # Maximum number of events to dequeue from the completion port on each pass
@@ -379,37 +296,146 @@ class _WindowsStatistics:
 MAX_EVENTS = 1000
 
 
+def _check(success: T) -> T:
+    if not success:
+        raise_winerror()
+    return success
+
+
+def _get_underlying_socket(
+    sock: _HasFileNo | int | Handle, *, which: WSAIoctls = WSAIoctls.SIO_BASE_HANDLE
+) -> Handle:
+    if hasattr(sock, "fileno"):
+        sock = sock.fileno()
+    base_ptr = ffi.new("HANDLE *")
+    out_size = ffi.new("DWORD *")
+    failed = ws2_32.WSAIoctl(
+        ffi.cast("SOCKET", sock),
+        which,
+        ffi.NULL,
+        0,
+        base_ptr,
+        ffi.sizeof("HANDLE"),
+        out_size,
+        ffi.NULL,
+        ffi.NULL,
+    )
+    if failed:
+        code = ws2_32.WSAGetLastError()
+        raise_winerror(code)
+    return Handle(base_ptr[0])
+
+
+def _get_base_socket(sock: _HasFileNo | int | Handle) -> Handle:
+    # There is a development kit for LSPs called Komodia Redirector.
+    # It does some unusual (some might say evil) things like intercepting
+    # SIO_BASE_HANDLE (fails) and SIO_BSP_HANDLE_SELECT (returns the same
+    # socket) in a misguided attempt to prevent bypassing it. It's been used
+    # in malware including the infamous Lenovo Superfish incident from 2015,
+    # but unfortunately is also used in some legitimate products such as
+    # parental control tools and Astrill VPN. Komodia happens to not
+    # block SIO_BSP_HANDLE_POLL, so we'll try SIO_BASE_HANDLE and fall back
+    # to SIO_BSP_HANDLE_POLL if it doesn't work.
+    # References:
+    # - https://github.com/piscisaureus/wepoll/blob/0598a791bf9cbbf480793d778930fc635b044980/wepoll.c#L2223
+    # - https://github.com/tokio-rs/mio/issues/1314
+
+    while True:
+        try:
+            # If this is not a Komodia-intercepted socket, we can just use
+            # SIO_BASE_HANDLE.
+            return _get_underlying_socket(sock)
+        except OSError as ex:
+            if ex.winerror == ErrorCodes.ERROR_NOT_SOCKET:
+                # SIO_BASE_HANDLE might fail even without LSP intervention,
+                # if we get something that's not a socket.
+                raise
+            if hasattr(sock, "fileno"):
+                sock = sock.fileno()
+            sock = _handle(sock)
+            next_sock = _get_underlying_socket(
+                sock, which=WSAIoctls.SIO_BSP_HANDLE_POLL
+            )
+            if next_sock == sock:
+                # If BSP_HANDLE_POLL returns the same socket we already had,
+                # then there's no layering going on and we need to fail
+                # to prevent an infinite loop.
+                raise RuntimeError(
+                    "Unexpected network configuration detected: "
+                    "SIO_BASE_HANDLE failed and SIO_BSP_HANDLE_POLL didn't "
+                    "return a different socket. Please file a bug at "
+                    "https://github.com/python-trio/trio/issues/new, "
+                    "and include the output of running: "
+                    "netsh winsock show catalog"
+                ) from ex
+            # Otherwise we've gotten at least one layer deeper, so
+            # loop back around to keep digging.
+            sock = next_sock
+
+
+def _afd_helper_handle() -> Handle:
+    # The "AFD" driver is exposed at the NT path "\Device\Afd". We're using
+    # the Win32 CreateFile, though, so we have to pass a Win32 path. \\.\ is
+    # how Win32 refers to the NT \GLOBAL??\ directory, and GLOBALROOT is a
+    # symlink inside that directory that points to the root of the NT path
+    # system. So by sticking that in front of the NT path, we get a Win32
+    # path. Alternatively, we could use NtCreateFile directly, since it takes
+    # an NT path. But we already wrap CreateFileW so this was easier.
+    # References:
+    #   https://blogs.msdn.microsoft.com/jeremykuhne/2016/05/02/dos-to-nt-a-paths-journey/
+    #   https://stackoverflow.com/a/21704022
+    #
+    # I'm actually not sure what the \Trio part at the end of the path does.
+    # Wepoll uses \Device\Afd\Wepoll, so I just copied them. (I'm guessing it
+    # might be visible in some debug tools, and is otherwise arbitrary?)
+    rawname = r"\\.\GLOBALROOT\Device\Afd\Trio".encode("utf-16le") + b"\0\0"
+    rawname_buf = ffi.from_buffer(rawname)
+
+    handle = kernel32.CreateFileW(
+        ffi.cast("LPCWSTR", rawname_buf),
+        FileFlags.SYNCHRONIZE,
+        FileFlags.FILE_SHARE_READ | FileFlags.FILE_SHARE_WRITE,
+        ffi.NULL,  # no security attributes
+        FileFlags.OPEN_EXISTING,
+        FileFlags.FILE_FLAG_OVERLAPPED,
+        ffi.NULL,  # no template file
+    )
+    if handle == INVALID_HANDLE_VALUE:  # pragma: no cover
+        raise_winerror()
+    return handle
+
+
 @attr.s(frozen=True)
 class CompletionKeyEventInfo:
-    lpOverlapped = attr.ib()
-    dwNumberOfBytesTransferred = attr.ib()
+    lpOverlapped: CData = attr.ib()
+    dwNumberOfBytesTransferred: int = attr.ib()
 
 
 class WindowsIOManager:
-    def __init__(self):
+    def __init__(self) -> None:
         # If this method raises an exception, then __del__ could run on a
         # half-initialized object. So we initialize everything that __del__
         # touches to safe values up front, before we do anything that can
         # fail.
         self._iocp = None
-        self._all_afd_handles = []
+        self._all_afd_handles: list[Handle] = []
 
         self._iocp = _check(
             kernel32.CreateIoCompletionPort(INVALID_HANDLE_VALUE, ffi.NULL, 0, 0)
         )
         self._events = ffi.new("OVERLAPPED_ENTRY[]", MAX_EVENTS)
 
-        self._vacant_afd_groups = set()
+        self._vacant_afd_groups: set[AFDGroup] = set()
         # {lpOverlapped: AFDPollOp}
-        self._afd_ops = {}
+        self._afd_ops: dict[CData, AFDPollOp] = {}
         # {socket handle: AFDWaiters}
-        self._afd_waiters = {}
+        self._afd_waiters: dict[Handle, AFDWaiters] = {}
 
         # {lpOverlapped: task}
-        self._overlapped_waiters = {}
-        self._posted_too_late_to_cancel = set()
+        self._overlapped_waiters: dict[CData, _core.Task] = {}
+        self._posted_too_late_to_cancel: set[CData] = set()
 
-        self._completion_key_queues = {}
+        self._completion_key_queues: dict[int, UnboundedQueue[object]] = {}
         self._completion_key_counter = itertools.count(CKeys.USER_DEFINED)
 
         with socket.socket() as s:
@@ -449,7 +475,7 @@ class WindowsIOManager:
                         "netsh winsock show catalog"
                     )
 
-    def close(self):
+    def close(self) -> None:
         try:
             if self._iocp is not None:
                 iocp = self._iocp
@@ -460,10 +486,10 @@ class WindowsIOManager:
                 afd_handle = self._all_afd_handles.pop()
                 _check(kernel32.CloseHandle(afd_handle))
 
-    def __del__(self):
+    def __del__(self) -> None:
         self.close()
 
-    def statistics(self):
+    def statistics(self) -> _WindowsStatistics:
         tasks_waiting_read = 0
         tasks_waiting_write = 0
         for waiter in self._afd_waiters.values():
@@ -478,19 +504,21 @@ class WindowsIOManager:
             completion_key_monitors=len(self._completion_key_queues),
         )
 
-    def force_wakeup(self):
+    def force_wakeup(self) -> None:
+        assert self._iocp is not None
         _check(
             kernel32.PostQueuedCompletionStatus(
                 self._iocp, 0, CKeys.FORCE_WAKEUP, ffi.NULL
             )
         )
 
-    def get_events(self, timeout):
+    def get_events(self, timeout: float) -> EventResult:
         received = ffi.new("PULONG")
         milliseconds = round(1000 * timeout)
         if timeout > 0 and milliseconds == 0:
             milliseconds = 1
         try:
+            assert self._iocp is not None
             _check(
                 kernel32.GetQueuedCompletionStatusEx(
                     self._iocp, self._events, MAX_EVENTS, received, milliseconds, 0
@@ -500,9 +528,11 @@ class WindowsIOManager:
             if exc.winerror != ErrorCodes.WAIT_TIMEOUT:  # pragma: no cover
                 raise
             return 0
-        return received[0]
+        result = received[0]
+        assert isinstance(result, int)
+        return result
 
-    def process_events(self, received):
+    def process_events(self, received: EventResult) -> None:
         for i in range(received):
             entry = self._events[i]
             if entry.lpCompletionKey == CKeys.AFD_POLL:
@@ -554,9 +584,9 @@ class WindowsIOManager:
                     pass
                 else:
                     exc = _core.TrioInternalError(
-                        "Failed to cancel overlapped I/O in {} and didn't "
+                        f"Failed to cancel overlapped I/O in {waiter.name} and didn't "
                         "receive the completion either. Did you forget to "
-                        "call register_with_iocp()?".format(waiter.name)
+                        "call register_with_iocp()?"
                     )
                     # Raising this out of handle_io ensures that
                     # the user will see our message even if some
@@ -582,8 +612,9 @@ class WindowsIOManager:
                 )
                 queue.put_nowait(info)
 
-    def _register_with_iocp(self, handle, completion_key):
-        handle = _handle(handle)
+    def _register_with_iocp(self, handle_: int | CData, completion_key: int) -> None:
+        handle = _handle(handle_)
+        assert self._iocp is not None
         _check(kernel32.CreateIoCompletionPort(handle, self._iocp, completion_key, 0))
         # Supposedly this makes things slightly faster, by disabling the
         # ability to do WaitForSingleObject(handle). We would never want to do
@@ -599,7 +630,7 @@ class WindowsIOManager:
     # AFD stuff
     ################################################################
 
-    def _refresh_afd(self, base_handle):
+    def _refresh_afd(self, base_handle: Handle) -> None:
         waiters = self._afd_waiters[base_handle]
         if waiters.current_op is not None:
             afd_group = waiters.current_op.afd_group
@@ -637,7 +668,7 @@ class WindowsIOManager:
 
             lpOverlapped = ffi.new("LPOVERLAPPED")
 
-            poll_info = ffi.new("AFD_POLL_INFO *")
+            poll_info: Any = ffi.new("AFD_POLL_INFO *")
             poll_info.Timeout = 2**63 - 1  # INT64_MAX
             poll_info.NumberOfHandles = 1
             poll_info.Exclusive = 0
@@ -675,7 +706,7 @@ class WindowsIOManager:
             if afd_group.size >= MAX_AFD_GROUP_SIZE:
                 self._vacant_afd_groups.remove(afd_group)
 
-    async def _afd_poll(self, sock, mode):
+    async def _afd_poll(self, sock: _HasFileNo | int, mode: str) -> None:
         base_handle = _get_base_socket(sock)
         waiters = self._afd_waiters.get(base_handle)
         if waiters is None:
@@ -688,7 +719,7 @@ class WindowsIOManager:
         # we let it escape.
         self._refresh_afd(base_handle)
 
-        def abort_fn(_):
+        def abort_fn(_: RaiseCancelT) -> Abort:
             setattr(waiters, mode, None)
             self._refresh_afd(base_handle)
             return _core.Abort.SUCCEEDED
@@ -696,15 +727,15 @@ class WindowsIOManager:
         await _core.wait_task_rescheduled(abort_fn)
 
     @_public
-    async def wait_readable(self, sock):
+    async def wait_readable(self, sock: _HasFileNo | int) -> None:
         await self._afd_poll(sock, "read_task")
 
     @_public
-    async def wait_writable(self, sock):
+    async def wait_writable(self, sock: _HasFileNo | int) -> None:
         await self._afd_poll(sock, "write_task")
 
     @_public
-    def notify_closing(self, handle):
+    def notify_closing(self, handle: Handle | int | _HasFileNo) -> None:
         handle = _get_base_socket(handle)
         waiters = self._afd_waiters.get(handle)
         if waiters is not None:
@@ -716,12 +747,14 @@ class WindowsIOManager:
     ################################################################
 
     @_public
-    def register_with_iocp(self, handle):
+    def register_with_iocp(self, handle: int | CData) -> None:
         self._register_with_iocp(handle, CKeys.WAIT_OVERLAPPED)
 
     @_public
-    async def wait_overlapped(self, handle, lpOverlapped):
-        handle = _handle(handle)
+    async def wait_overlapped(
+        self, handle_: int | CData, lpOverlapped: CData | int
+    ) -> object:
+        handle = _handle(handle_)
         if isinstance(lpOverlapped, int):
             lpOverlapped = ffi.cast("LPOVERLAPPED", lpOverlapped)
         if lpOverlapped in self._overlapped_waiters:
@@ -732,13 +765,14 @@ class WindowsIOManager:
         self._overlapped_waiters[lpOverlapped] = task
         raise_cancel = None
 
-        def abort(raise_cancel_):
+        def abort(raise_cancel_: RaiseCancelT) -> Abort:
             nonlocal raise_cancel
             raise_cancel = raise_cancel_
             try:
                 _check(kernel32.CancelIoEx(handle, lpOverlapped))
             except OSError as exc:
                 if exc.winerror == ErrorCodes.ERROR_NOT_FOUND:
+                    assert self._iocp is not None
                     # Too late to cancel. If this happens because the
                     # operation is already completed, we don't need to do
                     # anything; we'll get a notification of that completion
@@ -767,12 +801,14 @@ class WindowsIOManager:
                     ) from exc
             return _core.Abort.FAILED
 
+        # TODO: what type does this return?
         info = await _core.wait_task_rescheduled(abort)
-        if lpOverlapped.Internal != 0:
+        lpOverlappedTyped = cast("_Overlapped", lpOverlapped)
+        if lpOverlappedTyped.Internal != 0:
             # the lpOverlapped reports the error as an NT status code,
             # which we must convert back to a Win32 error code before
             # it will produce the right sorts of exceptions
-            code = ntdll.RtlNtStatusToDosError(lpOverlapped.Internal)
+            code = ntdll.RtlNtStatusToDosError(lpOverlappedTyped.Internal)
             if code == ErrorCodes.ERROR_OPERATION_ABORTED:
                 if raise_cancel is not None:
                     raise_cancel()
@@ -785,7 +821,9 @@ class WindowsIOManager:
                 raise_winerror(code)
         return info
 
-    async def _perform_overlapped(self, handle, submit_fn):
+    async def _perform_overlapped(
+        self, handle: int | CData, submit_fn: Callable[[_Overlapped], None]
+    ) -> _Overlapped:
         # submit_fn(lpOverlapped) submits some I/O
         # it may raise an OSError with ERROR_IO_PENDING
         # the handle must already be registered using
@@ -794,20 +832,22 @@ class WindowsIOManager:
         # operation will not be cancellable, depending on how Windows is
         # feeling today. So we need to check for cancellation manually.
         await _core.checkpoint_if_cancelled()
-        lpOverlapped = ffi.new("LPOVERLAPPED")
+        lpOverlapped = cast(_Overlapped, ffi.new("LPOVERLAPPED"))
         try:
             submit_fn(lpOverlapped)
         except OSError as exc:
             if exc.winerror != ErrorCodes.ERROR_IO_PENDING:
                 raise
-        await self.wait_overlapped(handle, lpOverlapped)
+        await self.wait_overlapped(handle, cast(CData, lpOverlapped))
         return lpOverlapped
 
     @_public
-    async def write_overlapped(self, handle, data, file_offset=0):
+    async def write_overlapped(
+        self, handle: int | CData, data: Buffer, file_offset: int = 0
+    ) -> int:
         with ffi.from_buffer(data) as cbuf:
 
-            def submit_write(lpOverlapped):
+            def submit_write(lpOverlapped: _Overlapped) -> None:
                 # yes, these are the real documented names
                 offset_fields = lpOverlapped.DUMMYUNIONNAME.DUMMYSTRUCTNAME
                 offset_fields.Offset = file_offset & 0xFFFFFFFF
@@ -827,10 +867,12 @@ class WindowsIOManager:
             return lpOverlapped.InternalHigh
 
     @_public
-    async def readinto_overlapped(self, handle, buffer, file_offset=0):
+    async def readinto_overlapped(
+        self, handle: int | CData, buffer: Buffer, file_offset: int = 0
+    ) -> int:
         with ffi.from_buffer(buffer, require_writable=True) as cbuf:
 
-            def submit_read(lpOverlapped):
+            def submit_read(lpOverlapped: _Overlapped) -> None:
                 offset_fields = lpOverlapped.DUMMYUNIONNAME.DUMMYSTRUCTNAME
                 offset_fields.Offset = file_offset & 0xFFFFFFFF
                 offset_fields.OffsetHigh = file_offset >> 32
@@ -852,14 +894,15 @@ class WindowsIOManager:
     ################################################################
 
     @_public
-    def current_iocp(self):
+    def current_iocp(self) -> int:
+        assert self._iocp is not None
         return int(ffi.cast("uintptr_t", self._iocp))
 
     @contextmanager
     @_public
-    def monitor_completion_key(self):
+    def monitor_completion_key(self) -> Iterator[tuple[int, UnboundedQueue[object]]]:
         key = next(self._completion_key_counter)
-        queue = _core.UnboundedQueue()
+        queue = _core.UnboundedQueue[object]()
         self._completion_key_queues[key] = queue
         try:
             yield (key, queue)
