@@ -1,21 +1,40 @@
+from __future__ import annotations
+
 import math
+from typing import TYPE_CHECKING, Protocol
 
 import attr
 
 import trio
 
 from . import _core
-from ._core import enable_ki_protection, ParkingLot
-from ._util import Final
+from ._core import Abort, ParkingLot, RaiseCancelT, enable_ki_protection
+from ._util import final
+
+if TYPE_CHECKING:
+    from types import TracebackType
+
+    from ._core import Task
+    from ._core._parking_lot import ParkingLotStatistics
 
 
-@attr.s(frozen=True)
-class _EventStatistics:
-    tasks_waiting = attr.ib()
+@attr.s(frozen=True, slots=True)
+class EventStatistics:
+    """An object containing debugging information.
+
+    Currently the following fields are defined:
+
+    * ``tasks_waiting``: The number of tasks blocked on this event's
+      :meth:`trio.Event.wait` method.
+
+    """
+
+    tasks_waiting: int = attr.ib()
 
 
+@final
 @attr.s(repr=False, eq=False, hash=False, slots=True)
-class Event(metaclass=Final):
+class Event:
     """A waitable boolean value useful for inter-task synchronization,
     inspired by :class:`threading.Event`.
 
@@ -41,15 +60,15 @@ class Event(metaclass=Final):
 
     """
 
-    _tasks = attr.ib(factory=set, init=False)
-    _flag = attr.ib(default=False, init=False)
+    _tasks: set[Task] = attr.ib(factory=set, init=False)
+    _flag: bool = attr.ib(default=False, init=False)
 
-    def is_set(self):
+    def is_set(self) -> bool:
         """Return the current value of the internal flag."""
         return self._flag
 
     @enable_ki_protection
-    def set(self):
+    def set(self) -> None:
         """Set the internal flag value to True, and wake any waiting tasks."""
         if not self._flag:
             self._flag = True
@@ -57,7 +76,7 @@ class Event(metaclass=Final):
                 _core.reschedule(task)
             self._tasks.clear()
 
-    async def wait(self):
+    async def wait(self) -> None:
         """Block until the internal flag value becomes True.
 
         If it's already True, then this method returns immediately.
@@ -69,13 +88,13 @@ class Event(metaclass=Final):
             task = _core.current_task()
             self._tasks.add(task)
 
-            def abort_fn(_):
+            def abort_fn(_: RaiseCancelT) -> Abort:
                 self._tasks.remove(task)
                 return _core.Abort.SUCCEEDED
 
             await _core.wait_task_rescheduled(abort_fn)
 
-    def statistics(self):
+    def statistics(self) -> EventStatistics:
         """Return an object containing debugging information.
 
         Currently the following fields are defined:
@@ -84,36 +103,64 @@ class Event(metaclass=Final):
           :meth:`wait` method.
 
         """
-        return _EventStatistics(tasks_waiting=len(self._tasks))
+        return EventStatistics(tasks_waiting=len(self._tasks))
 
 
-def async_cm(cls):
+class _HasAcquireRelease(Protocol):
+    """Only classes with acquire() and release() can use the mixin's implementations."""
+
+    async def acquire(self) -> object:
+        ...
+
+    def release(self) -> object:
+        ...
+
+
+class AsyncContextManagerMixin:
     @enable_ki_protection
-    async def __aenter__(self):
+    async def __aenter__(self: _HasAcquireRelease) -> None:
         await self.acquire()
 
-    __aenter__.__qualname__ = cls.__qualname__ + ".__aenter__"
-    cls.__aenter__ = __aenter__
-
     @enable_ki_protection
-    async def __aexit__(self, *args):
+    async def __aexit__(
+        self: _HasAcquireRelease,
+        exc_type: type[BaseException] | None,
+        exc_value: BaseException | None,
+        traceback: TracebackType | None,
+    ) -> None:
         self.release()
 
-    __aexit__.__qualname__ = cls.__qualname__ + ".__aexit__"
-    cls.__aexit__ = __aexit__
-    return cls
+
+@attr.s(frozen=True, slots=True)
+class CapacityLimiterStatistics:
+    """An object containing debugging information.
+
+    Currently the following fields are defined:
+
+    * ``borrowed_tokens``: The number of tokens currently borrowed from
+      the sack.
+    * ``total_tokens``: The total number of tokens in the sack. Usually
+      this will be larger than ``borrowed_tokens``, but it's possibly for
+      it to be smaller if :attr:`trio.CapacityLimiter.total_tokens` was recently decreased.
+    * ``borrowers``: A list of all tasks or other entities that currently
+      hold a token.
+    * ``tasks_waiting``: The number of tasks blocked on this
+      :class:`CapacityLimiter`\'s :meth:`trio.CapacityLimiter.acquire` or
+      :meth:`trio.CapacityLimiter.acquire_on_behalf_of` methods.
+
+    """
+
+    borrowed_tokens: int = attr.ib()
+    total_tokens: int | float = attr.ib()
+    borrowers: list[Task | object] = attr.ib()
+    tasks_waiting: int = attr.ib()
 
 
-@attr.s(frozen=True)
-class _CapacityLimiterStatistics:
-    borrowed_tokens = attr.ib()
-    total_tokens = attr.ib()
-    borrowers = attr.ib()
-    tasks_waiting = attr.ib()
-
-
-@async_cm
-class CapacityLimiter(metaclass=Final):
+# Can be a generic type with a default of Task if/when PEP 696 is released
+# and implemented in type checkers. Making it fully generic would currently
+# introduce a lot of unnecessary hassle.
+@final
+class CapacityLimiter(AsyncContextManagerMixin):
     """An object for controlling access to a resource with limited capacity.
 
     Sometimes you need to put a limit on how many tasks can do something at
@@ -167,22 +214,23 @@ class CapacityLimiter(metaclass=Final):
 
     """
 
-    def __init__(self, total_tokens):
+    # total_tokens would ideally be int|Literal[math.inf] - but that's not valid typing
+    def __init__(self, total_tokens: int | float):  # noqa: PYI041
         self._lot = ParkingLot()
-        self._borrowers = set()
+        self._borrowers: set[Task | object] = set()
         # Maps tasks attempting to acquire -> borrower, to handle on-behalf-of
-        self._pending_borrowers = {}
+        self._pending_borrowers: dict[Task, Task | object] = {}
         # invoke the property setter for validation
-        self.total_tokens = total_tokens
+        self.total_tokens: int | float = total_tokens
         assert self._total_tokens == total_tokens
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "<trio.CapacityLimiter at {:#x}, {}/{} with {} waiting>".format(
             id(self), len(self._borrowers), self._total_tokens, len(self._lot)
         )
 
     @property
-    def total_tokens(self):
+    def total_tokens(self) -> int | float:
         """The total capacity available.
 
         You can change :attr:`total_tokens` by assigning to this attribute. If
@@ -197,7 +245,7 @@ class CapacityLimiter(metaclass=Final):
         return self._total_tokens
 
     @total_tokens.setter
-    def total_tokens(self, new_total_tokens):
+    def total_tokens(self, new_total_tokens: int | float) -> None:  # noqa: PYI041
         if not isinstance(new_total_tokens, int) and new_total_tokens != math.inf:
             raise TypeError("total_tokens must be an int or math.inf")
         if new_total_tokens < 1:
@@ -205,23 +253,23 @@ class CapacityLimiter(metaclass=Final):
         self._total_tokens = new_total_tokens
         self._wake_waiters()
 
-    def _wake_waiters(self):
+    def _wake_waiters(self) -> None:
         available = self._total_tokens - len(self._borrowers)
         for woken in self._lot.unpark(count=available):
             self._borrowers.add(self._pending_borrowers.pop(woken))
 
     @property
-    def borrowed_tokens(self):
+    def borrowed_tokens(self) -> int:
         """The amount of capacity that's currently in use."""
         return len(self._borrowers)
 
     @property
-    def available_tokens(self):
+    def available_tokens(self) -> int | float:
         """The amount of capacity that's available to use."""
         return self.total_tokens - self.borrowed_tokens
 
     @enable_ki_protection
-    def acquire_nowait(self):
+    def acquire_nowait(self) -> None:
         """Borrow a token from the sack, without blocking.
 
         Raises:
@@ -233,7 +281,7 @@ class CapacityLimiter(metaclass=Final):
         self.acquire_on_behalf_of_nowait(trio.lowlevel.current_task())
 
     @enable_ki_protection
-    def acquire_on_behalf_of_nowait(self, borrower):
+    def acquire_on_behalf_of_nowait(self, borrower: Task | object) -> None:
         """Borrow a token from the sack on behalf of ``borrower``, without
         blocking.
 
@@ -253,8 +301,7 @@ class CapacityLimiter(metaclass=Final):
         """
         if borrower in self._borrowers:
             raise RuntimeError(
-                "this borrower is already holding one of this "
-                "CapacityLimiter's tokens"
+                "this borrower is already holding one of this CapacityLimiter's tokens"
             )
         if len(self._borrowers) < self._total_tokens and not self._lot:
             self._borrowers.add(borrower)
@@ -262,7 +309,7 @@ class CapacityLimiter(metaclass=Final):
             raise trio.WouldBlock
 
     @enable_ki_protection
-    async def acquire(self):
+    async def acquire(self) -> None:
         """Borrow a token from the sack, blocking if necessary.
 
         Raises:
@@ -273,7 +320,7 @@ class CapacityLimiter(metaclass=Final):
         await self.acquire_on_behalf_of(trio.lowlevel.current_task())
 
     @enable_ki_protection
-    async def acquire_on_behalf_of(self, borrower):
+    async def acquire_on_behalf_of(self, borrower: Task | object) -> None:
         """Borrow a token from the sack on behalf of ``borrower``, blocking if
         necessary.
 
@@ -302,7 +349,7 @@ class CapacityLimiter(metaclass=Final):
             await trio.lowlevel.cancel_shielded_checkpoint()
 
     @enable_ki_protection
-    def release(self):
+    def release(self) -> None:
         """Put a token back into the sack.
 
         Raises:
@@ -313,7 +360,7 @@ class CapacityLimiter(metaclass=Final):
         self.release_on_behalf_of(trio.lowlevel.current_task())
 
     @enable_ki_protection
-    def release_on_behalf_of(self, borrower):
+    def release_on_behalf_of(self, borrower: Task | object) -> None:
         """Put a token back into the sack on behalf of ``borrower``.
 
         Raises:
@@ -328,7 +375,7 @@ class CapacityLimiter(metaclass=Final):
         self._borrowers.remove(borrower)
         self._wake_waiters()
 
-    def statistics(self):
+    def statistics(self) -> CapacityLimiterStatistics:
         """Return an object containing debugging information.
 
         Currently the following fields are defined:
@@ -345,7 +392,7 @@ class CapacityLimiter(metaclass=Final):
           :meth:`acquire_on_behalf_of` methods.
 
         """
-        return _CapacityLimiterStatistics(
+        return CapacityLimiterStatistics(
             borrowed_tokens=len(self._borrowers),
             total_tokens=self._total_tokens,
             # Use a list instead of a frozenset just in case we start to allow
@@ -355,8 +402,8 @@ class CapacityLimiter(metaclass=Final):
         )
 
 
-@async_cm
-class Semaphore(metaclass=Final):
+@final
+class Semaphore(AsyncContextManagerMixin):
     """A `semaphore <https://en.wikipedia.org/wiki/Semaphore_(programming)>`__.
 
     A semaphore holds an integer value, which can be incremented by
@@ -383,7 +430,7 @@ class Semaphore(metaclass=Final):
 
     """
 
-    def __init__(self, initial_value, *, max_value=None):
+    def __init__(self, initial_value: int, *, max_value: int | None = None):
         if not isinstance(initial_value, int):
             raise TypeError("initial_value must be an int")
         if initial_value < 0:
@@ -401,27 +448,25 @@ class Semaphore(metaclass=Final):
         self._value = initial_value
         self._max_value = max_value
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         if self._max_value is None:
             max_value_str = ""
         else:
-            max_value_str = ", max_value={}".format(self._max_value)
-        return "<trio.Semaphore({}{}) at {:#x}>".format(
-            self._value, max_value_str, id(self)
-        )
+            max_value_str = f", max_value={self._max_value}"
+        return f"<trio.Semaphore({self._value}{max_value_str}) at {id(self):#x}>"
 
     @property
-    def value(self):
+    def value(self) -> int:
         """The current value of the semaphore."""
         return self._value
 
     @property
-    def max_value(self):
+    def max_value(self) -> int | None:
         """The maximum allowed value. May be None to indicate no limit."""
         return self._max_value
 
     @enable_ki_protection
-    def acquire_nowait(self):
+    def acquire_nowait(self) -> None:
         """Attempt to decrement the semaphore value, without blocking.
 
         Raises:
@@ -435,7 +480,7 @@ class Semaphore(metaclass=Final):
             raise trio.WouldBlock
 
     @enable_ki_protection
-    async def acquire(self):
+    async def acquire(self) -> None:
         """Decrement the semaphore value, blocking if necessary to avoid
         letting it drop below zero.
 
@@ -449,7 +494,7 @@ class Semaphore(metaclass=Final):
             await trio.lowlevel.cancel_shielded_checkpoint()
 
     @enable_ki_protection
-    def release(self):
+    def release(self) -> None:
         """Increment the semaphore value, possibly waking a task blocked in
         :meth:`acquire`.
 
@@ -466,7 +511,7 @@ class Semaphore(metaclass=Final):
                 raise ValueError("semaphore released too many times")
             self._value += 1
 
-    def statistics(self):
+    def statistics(self) -> ParkingLotStatistics:
         """Return an object containing debugging information.
 
         Currently the following fields are defined:
@@ -478,31 +523,40 @@ class Semaphore(metaclass=Final):
         return self._lot.statistics()
 
 
-@attr.s(frozen=True)
-class _LockStatistics:
-    locked = attr.ib()
-    owner = attr.ib()
-    tasks_waiting = attr.ib()
+@attr.s(frozen=True, slots=True)
+class LockStatistics:
+    """An object containing debugging information for a Lock.
+
+    Currently the following fields are defined:
+
+    * ``locked`` (boolean): indicating whether the lock is held.
+    * ``owner``: the :class:`trio.lowlevel.Task` currently holding the lock,
+      or None if the lock is not held.
+    * ``tasks_waiting`` (int): The number of tasks blocked on this lock's
+      :meth:`trio.Lock.acquire` method.
+
+    """
+
+    locked: bool = attr.ib()
+    owner: Task | None = attr.ib()
+    tasks_waiting: int = attr.ib()
 
 
-@async_cm
 @attr.s(eq=False, hash=False, repr=False)
-class _LockImpl:
-    _lot = attr.ib(factory=ParkingLot, init=False)
-    _owner = attr.ib(default=None, init=False)
+class _LockImpl(AsyncContextManagerMixin):
+    _lot: ParkingLot = attr.ib(factory=ParkingLot, init=False)
+    _owner: Task | None = attr.ib(default=None, init=False)
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         if self.locked():
             s1 = "locked"
-            s2 = " with {} waiters".format(len(self._lot))
+            s2 = f" with {len(self._lot)} waiters"
         else:
             s1 = "unlocked"
             s2 = ""
-        return "<{} {} object at {:#x}{}>".format(
-            s1, self.__class__.__name__, id(self), s2
-        )
+        return f"<{s1} {self.__class__.__name__} object at {id(self):#x}{s2}>"
 
-    def locked(self):
+    def locked(self) -> bool:
         """Check whether the lock is currently held.
 
         Returns:
@@ -512,7 +566,7 @@ class _LockImpl:
         return self._owner is not None
 
     @enable_ki_protection
-    def acquire_nowait(self):
+    def acquire_nowait(self) -> None:
         """Attempt to acquire the lock, without blocking.
 
         Raises:
@@ -530,7 +584,7 @@ class _LockImpl:
             raise trio.WouldBlock
 
     @enable_ki_protection
-    async def acquire(self):
+    async def acquire(self) -> None:
         """Acquire the lock, blocking if necessary."""
         await trio.lowlevel.checkpoint_if_cancelled()
         try:
@@ -544,7 +598,7 @@ class _LockImpl:
             await trio.lowlevel.cancel_shielded_checkpoint()
 
     @enable_ki_protection
-    def release(self):
+    def release(self) -> None:
         """Release the lock.
 
         Raises:
@@ -559,7 +613,7 @@ class _LockImpl:
         else:
             self._owner = None
 
-    def statistics(self):
+    def statistics(self) -> LockStatistics:
         """Return an object containing debugging information.
 
         Currently the following fields are defined:
@@ -571,12 +625,13 @@ class _LockImpl:
           :meth:`acquire` method.
 
         """
-        return _LockStatistics(
+        return LockStatistics(
             locked=self.locked(), owner=self._owner, tasks_waiting=len(self._lot)
         )
 
 
-class Lock(_LockImpl, metaclass=Final):
+@final
+class Lock(_LockImpl):
     """A classic `mutex
     <https://en.wikipedia.org/wiki/Lock_(computer_science)>`__.
 
@@ -590,7 +645,8 @@ class Lock(_LockImpl, metaclass=Final):
     """
 
 
-class StrictFIFOLock(_LockImpl, metaclass=Final):
+@final
+class StrictFIFOLock(_LockImpl):
     r"""A variant of :class:`Lock` where tasks are guaranteed to acquire the
     lock in strict first-come-first-served order.
 
@@ -653,14 +709,24 @@ class StrictFIFOLock(_LockImpl, metaclass=Final):
     """
 
 
-@attr.s(frozen=True)
-class _ConditionStatistics:
-    tasks_waiting = attr.ib()
-    lock_statistics = attr.ib()
+@attr.s(frozen=True, slots=True)
+class ConditionStatistics:
+    r"""An object containing debugging information for a Condition.
+
+    Currently the following fields are defined:
+
+    * ``tasks_waiting`` (int): The number of tasks blocked on this condition's
+      :meth:`trio.Condition.wait` method.
+    * ``lock_statistics``: The result of calling the underlying
+      :class:`Lock`\s  :meth:`~Lock.statistics` method.
+
+    """
+    tasks_waiting: int = attr.ib()
+    lock_statistics: LockStatistics = attr.ib()
 
 
-@async_cm
-class Condition(metaclass=Final):
+@final
+class Condition(AsyncContextManagerMixin):
     """A classic `condition variable
     <https://en.wikipedia.org/wiki/Monitor_(synchronization)>`__, similar to
     :class:`threading.Condition`.
@@ -675,15 +741,15 @@ class Condition(metaclass=Final):
 
     """
 
-    def __init__(self, lock=None):
+    def __init__(self, lock: Lock | None = None):
         if lock is None:
             lock = Lock()
-        if not type(lock) is Lock:
+        if type(lock) is not Lock:
             raise TypeError("lock must be a trio.Lock")
         self._lock = lock
         self._lot = trio.lowlevel.ParkingLot()
 
-    def locked(self):
+    def locked(self) -> bool:
         """Check whether the underlying lock is currently held.
 
         Returns:
@@ -692,7 +758,7 @@ class Condition(metaclass=Final):
         """
         return self._lock.locked()
 
-    def acquire_nowait(self):
+    def acquire_nowait(self) -> None:
         """Attempt to acquire the underlying lock, without blocking.
 
         Raises:
@@ -701,16 +767,16 @@ class Condition(metaclass=Final):
         """
         return self._lock.acquire_nowait()
 
-    async def acquire(self):
+    async def acquire(self) -> None:
         """Acquire the underlying lock, blocking if necessary."""
         await self._lock.acquire()
 
-    def release(self):
+    def release(self) -> None:
         """Release the underlying lock."""
         self._lock.release()
 
     @enable_ki_protection
-    async def wait(self):
+    async def wait(self) -> None:
         """Wait for another task to call :meth:`notify` or
         :meth:`notify_all`.
 
@@ -745,7 +811,7 @@ class Condition(metaclass=Final):
                 await self.acquire()
             raise
 
-    def notify(self, n=1):
+    def notify(self, n: int = 1) -> None:
         """Wake one or more tasks that are blocked in :meth:`wait`.
 
         Args:
@@ -759,7 +825,7 @@ class Condition(metaclass=Final):
             raise RuntimeError("must hold the lock to notify")
         self._lot.repark(self._lock._lot, count=n)
 
-    def notify_all(self):
+    def notify_all(self) -> None:
         """Wake all tasks that are currently blocked in :meth:`wait`.
 
         Raises:
@@ -770,7 +836,7 @@ class Condition(metaclass=Final):
             raise RuntimeError("must hold the lock to notify")
         self._lot.repark_all(self._lock._lot)
 
-    def statistics(self):
+    def statistics(self) -> ConditionStatistics:
         r"""Return an object containing debugging information.
 
         Currently the following fields are defined:
@@ -781,6 +847,6 @@ class Condition(metaclass=Final):
           :class:`Lock`\s  :meth:`~Lock.statistics` method.
 
         """
-        return _ConditionStatistics(
+        return ConditionStatistics(
             tasks_waiting=len(self._lot), lock_statistics=self._lock.statistics()
         )
